/*
 * Copyright (C) 2008  Internet Systems Consortium, Inc. ("ISC")
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/* $Id: selecttest.c,v 1.8 2008/09/20 22:34:53 jinmei Exp $ */

#ifdef TEST_FDSETSIZE
#define FD_SETSIZE TEST_FDSETSIZE
#else
#define TEST_FDSETSIZE -1	/* meaning 'undefined' */
#endif

#include <sys/types.h>
#include <sys/time.h>	/* Required on some systems for <sys/resource.h>. */
#include <sys/resource.h>
#include <sys/param.h>
#include <sys/socket.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#include <sys/errno.h>

#ifdef HAVE_SYS_PRCTL_H
#include <sys/prctl.h>		/* Required for prctl(). */

/*
 * If the value of PR_SET_KEEPCAPS is not in <sys/prctl.h>, define it
 * here.  This allows setuid() to work on systems running a new enough
 * kernel but with /usr/include/linux pointing to "standard" kernel
 * headers.
 */
#ifndef PR_SET_KEEPCAPS
#define PR_SET_KEEPCAPS 8
#endif
#endif

#include <netinet/in.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>

#ifdef __linux__
#include <linux/fs.h>	/* To get the large NR_OPEN. */
#endif

#ifdef __hpux
#include <sys/dyntune.h>
#endif

#define NSOCKS 4093		/* 4096 - stdin/out/err */

#ifdef RLIMIT_NOFILE
static void
set_nofile()
{
	struct rlimit rl;
	int ret;

	rl.rlim_cur = rl.rlim_max = RLIM_INFINITY;
	ret = setrlimit(RLIMIT_NOFILE, &rl);

#if defined(OPEN_MAX) && defined(__APPLE__)
	if (ret < 0) {
		rl.rlim_cur = OPEN_MAX;
		ret = setrlimit(RLIMIT_NOFILE, &rl);
	}
#elif defined(NR_OPEN) && defined(__linux__)
	if (ret < 0) {
		rl.rlim_cur = rl.rlim_max = NR_OPEN;
		ret = setrlimit(RLIMIT_NOFILE, &rl);
	}
#elif defined(__hpux)
	if (ret < 0) {
		uint64_t maxfiles;

		if (gettune("maxfiles_lim", &maxfiles) == 0) {
			rl.rlim_cur = rl.rlim_max = maxfiles;
			ret = setrlimit(RLIMIT_NOFILE, &rl);
		}
	}
#endif
	if (ret < 0) {
		perror("set resource limit");
		exit(1);
	}
}
#endif

int
main(int argc, char *argv[])
{
	int nsocks = NSOCKS;
	int set_resource = 0;
	int ch, n, i, s, s2, maxfd, error;
	char *user = NULL;
	fd_set *read_fds;
	size_t bufsize;
	struct addrinfo hints, *res;
	struct sockaddr_in sin;
	int sin_len;	/* using int would be safer in terms of portability */
	char buf[8];

	printf("selecttest: nsocks = %d, TEST_FDSETSIZE = %d, FD_SETSIZE = %d, "
	       "sizeof fd_set = %d\n", nsocks,
	       TEST_FDSETSIZE, FD_SETSIZE, sizeof(fd_set));

	while ((ch = getopt(argc, argv, "ru:")) != -1) {
		switch (ch) {
		case 'r':
			set_resource = 1;
			break;
		case 'u':
			user = optarg;
			break;
		default:
			fprintf(stderr, "usage: selecttest [-r] [-u username] "
				"[num_sockets]\n");
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (argc > 0)
		nsocks = atoi(argv[0]);

	if (user != NULL) {
		struct passwd *pw;

		pw = getpwnam(user);
		if (pw == NULL) {
			fprintf(stderr, "failed to find passwd entry for %s",
				user);
			exit(1);
		}

#ifdef HAVE_SYS_PRCTL_H
		if (prctl(PR_SET_KEEPCAPS, 1, 0, 0, 0) < 0) {
			perror("prctl(SET_KEEPCAPS)");
			exit(1);
		}
#endif
		if (setuid(pw->pw_uid) < 0) {
			perror("setuid");
			exit(1);
		}
	}

	if (set_resource) {
#ifdef RLIMIT_NOFILE
		set_nofile();
#else
		fprintf(stderr,
			"this system doesn't allow setting # of open files\n");
		exit(1);
#endif
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = IPPROTO_UDP;
	hints.ai_flags = AI_NUMERICHOST;
	error = getaddrinfo("127.0.0.1", NULL, &hints, &res);
	if (error != 0) {
		fprintf(stderr, "getaddrinfo failed: %s", gai_strerror(error));
		exit(1);
	}

	maxfd = -1;
	for (i = 0; i < nsocks; i++) {
		s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (s < 0) {
			fprintf(stderr, "opening %dth socket failed: %s\n",
				i + 1, strerror(errno));
			exit(1);
		}
		if (s > maxfd)
			maxfd = s;
		if (bind(s, res->ai_addr, res->ai_addrlen) < 0) {
			fprintf(stderr, "binding %dth socket (fd: %d) "
				"failed: %s\n", i + 1, s, strerror(errno));
			exit(1);
		}
	}
	printf("created %d sockets, maxfd = %d\n", nsocks, maxfd);

	bufsize = howmany(maxfd + 1, NFDBITS) * sizeof(fd_mask);
	read_fds = malloc(bufsize);
	if (read_fds == NULL) {
		perror("memory allocation for fd sets");
		exit(1);
	}

	memset(read_fds, 0xf0, bufsize); /* filled with dummy data */
	printf("FD_CLR test...");
	for (i = 0; i <= maxfd; i++)
		FD_CLR(i, read_fds);
	for (i = 0; i <= maxfd; i++) {
		if (FD_ISSET(i, read_fds)) {
			fprintf(stderr, "FD %d is not cleared\n", i);
			exit(1);
		}
	}
	printf("OK\n");

	memset(read_fds, 0xf0, bufsize); /* filled with dummy data */
	printf("FD_SET test...");
	for (i = 0; i <= maxfd; i++)
		FD_SET(i, read_fds);
	for (i = 0; i <= maxfd; i++) {
		if (!FD_ISSET(i, read_fds)) {
			fprintf(stderr, "FD %d is not set\n", i);
			exit(1);
		}
	}
	printf("OK\n");

	memset(read_fds, 0, bufsize);
	FD_SET(maxfd, read_fds);

	sin_len = sizeof(sin);
	if (getsockname(maxfd, (struct sockaddr *)&sin, (void *)&sin_len) < 0) {
		perror("getsockname");
		exit(1);
	}
	n = sendto(maxfd, buf, sizeof(buf), 0, (struct sockaddr *)&sin,
		   sin_len);
	if (n < 0) {
		perror("sendto");
		exit(1);
	} else if (n != sizeof(buf))
		fprintf(stderr, "unexpected sendto result: %d\n", n);

	printf("select test...");
	n = select(maxfd + 1, read_fds, NULL, NULL, NULL);
	if (n < 0) {
		perror("select");
		exit(1);
	} else if (n != 1)
		fprintf(stderr, "unexpected select result: %d\n", n);

	error = 0;
	for (i = 0; i <= maxfd; i++) {
		if (i != maxfd && FD_ISSET(i, read_fds)) {
			fprintf(stderr, "FD %d is unexpectedly set\n", i);
			error = 1;
		}
		if (i == maxfd && !FD_ISSET(i, read_fds)) {
			fprintf(stderr, "FD %d should be set but isn't\n", i);
			error = 1;
		}
	}

	if (error == 0)
		printf("OK\n");
	else {
		printf("NG\n");
		exit(1);
	}

	exit (0);
}
